@@ -1,11 +1,18 @@
+import eth_account
 import logging
+import secrets
 
+from eth_abi import encode
 from eth_account.signers.local import LocalAccount
+from eth_utils import keccak, to_hex
 
 from hyperliquid.api import API
 from hyperliquid.info import Info
+from hyperliquid.utils.constants import MAINNET_API_URL
 from hyperliquid.utils.signing import (
     ZERO_ADDRESS,
+    CancelRequest,
+    OrderRequest,
     OrderSpec,
     OrderType,
     float_to_usd_int,
@@ -14,8 +21,10 @@ from hyperliquid.utils.signing import (
     order_spec_preprocessing,
     order_spec_to_order_wire,
     sign_l1_action,
+    sign_usd_transfer_action,
+    sign_agent,
 )
-from hyperliquid.utils.types import Any, Literal, Meta, Optional
+from hyperliquid.utils.types import Any, List, Literal, Meta, Optional, Tuple
 
 
 class Exchange(API):
@@ -49,55 +58,78 @@ class Exchange(API):
     def order(
         self, coin: str, is_buy: bool, sz: float, limit_px: float, order_type: OrderType, reduce_only: bool = False
     ) -> Any:
-        order_spec: OrderSpec = {
-            "order": {
-                "asset": self.coin_to_asset[coin],
-                "isBuy": is_buy,
-                "reduceOnly": reduce_only,
-                "limitPx": limit_px,
-                "sz": sz,
-            },
-            "orderType": order_type,
-        }
+        return self.bulk_orders(
+            [
+                {
+                    "coin": coin,
+                    "is_buy": is_buy,
+                    "sz": sz,
+                    "limit_px": limit_px,
+                    "order_type": order_type,
+                    "reduce_only": reduce_only,
+                }
+            ]
+        )
+
+    def bulk_orders(self, order_requests: List[OrderRequest]) -> Any:
+        order_specs: List[OrderSpec] = [
+            {
+                "order": {
+                    "asset": self.coin_to_asset[order["coin"]],
+                    "isBuy": order["is_buy"],
+                    "reduceOnly": order["reduce_only"],
+                    "limitPx": order["limit_px"],
+                    "sz": order["sz"],
+                },
+                "orderType": order["order_type"],
+            }
+            for order in order_requests
+        ]
+
         timestamp = get_timestamp_ms()
         grouping: Literal["na"] = "na"
 
         signature = sign_l1_action(
             self.wallet,
             ["(uint32,bool,uint64,uint64,bool,uint8,uint64)[]", "uint8"],
-            [[order_spec_preprocessing(order_spec)], order_grouping_to_number(grouping)],
+            [[order_spec_preprocessing(order_spec) for order_spec in order_specs], order_grouping_to_number(grouping)],
             ZERO_ADDRESS if self.vault_address is None else self.vault_address,
             timestamp,
+            self.base_url == MAINNET_API_URL,
         )
 
         return self._post_action(
             {
                 "type": "order",
                 "grouping": grouping,
-                "orders": [order_spec_to_order_wire(order_spec)],
+                "orders": [order_spec_to_order_wire(order_spec) for order_spec in order_specs],
             },
             signature,
             timestamp,
         )
 
     def cancel(self, coin: str, oid: int) -> Any:
+        return self.bulk_cancel([{"coin": coin, "oid": oid}])
+
+    def bulk_cancel(self, cancel_requests: List[CancelRequest]) -> Any:
         timestamp = get_timestamp_ms()
-        asset = self.coin_to_asset[coin]
         signature = sign_l1_action(
             self.wallet,
             ["(uint32,uint64)[]"],
-            [[(asset, oid)]],
+            [[(self.coin_to_asset[cancel["coin"]], cancel["oid"]) for cancel in cancel_requests]],
             ZERO_ADDRESS if self.vault_address is None else self.vault_address,
             timestamp,
+            self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
             {
                 "type": "cancel",
                 "cancels": [
                     {
-                        "asset": asset,
-                        "oid": oid,
+                        "asset": self.coin_to_asset[cancel["coin"]],
+                        "oid": cancel["oid"],
                     }
+                    for cancel in cancel_requests
                 ],
             },
             signature,
@@ -113,6 +145,7 @@ class Exchange(API):
             [asset, is_cross, leverage],
             ZERO_ADDRESS if self.vault_address is None else self.vault_address,
             timestamp,
+            self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
             {
@@ -135,6 +168,7 @@ class Exchange(API):
             [asset, True, amount],
             ZERO_ADDRESS if self.vault_address is None else self.vault_address,
             timestamp,
+            self.base_url == MAINNET_API_URL,
         )
         return self._post_action(
             {
@@ -146,3 +180,47 @@ class Exchange(API):
             signature,
             timestamp,
         )
+
+    def usd_tranfer(self, amount: float, destination: str) -> Any:
+        timestamp = get_timestamp_ms()
+        payload = {
+            "destination": destination,
+            "amount": str(amount),
+            "time": timestamp,
+        }
+        is_mainnet = self.base_url == MAINNET_API_URL
+        signature = sign_usd_transfer_action(self.wallet, payload, is_mainnet)
+        return self._post_action(
+            {
+                "chain": "Arbitrum" if is_mainnet else "ArbitrumGoerli",
+                "payload": payload,
+                "type": "usdTransfer",
+            },
+            signature,
+            timestamp,
+        )
+
+    def approve_agent(self) -> Tuple[Any, str]:
+        agent_key = "0x" + secrets.token_hex(32)
+        account = eth_account.Account.from_key(agent_key)
+        agent = {
+            "source": "https://hyperliquid.xyz",
+            "connectionId": keccak(encode(["address"], [account.address])),
+        }
+        timestamp = get_timestamp_ms()
+        is_mainnet = self.base_url == MAINNET_API_URL
+        signature = sign_agent(self.wallet, agent, is_mainnet)
+        agent["connectionId"] = to_hex(agent["connectionId"])
+        return (
+            self._post_action(
+                {
+                    "chain": "Arbitrum" if is_mainnet else "ArbitrumGoerli",
+                    "agent": agent,
+                    "agentAddress": account.address,
+                    "type": "connect",
+                },
+                signature,
+                timestamp,
+            ),
+            agent_key,
+        )
