@@ -15,6 +15,11 @@ TriggerOrderType = TypedDict("TriggerOrderType", {"triggerPx": float, "isMarket"
 TriggerOrderTypeWire = TypedDict("TriggerOrderTypeWire", {"triggerPx": str, "isMarket": bool, "tpsl": Tpsl})
 OrderType = TypedDict("OrderType", {"limit": LimitOrderType, "trigger": TriggerOrderType}, total=False)
 OrderTypeWire = TypedDict("OrderTypeWire", {"limit": LimitOrderType, "trigger": TriggerOrderTypeWire}, total=False)
+OrderRequest = TypedDict(
+    "OrderRequest",
+    {"coin": str, "is_buy": bool, "sz": float, "limit_px": float, "order_type": OrderType, "reduce_only": bool},
+)
+CancelRequest = TypedDict("CancelRequest", {"coin": str, "oid": int})
 
 
 def order_type_to_tuple(order_type: OrderType) -> Tuple[int, float]:
@@ -102,13 +107,13 @@ def order_spec_to_order_wire(order_spec: OrderSpec) -> OrderWire:
     }
 
 
-def construct_phantom_agent(signature_types, signature_data):
+def construct_phantom_agent(signature_types, signature_data, is_mainnet):
     connection_id = encode(signature_types, signature_data)
 
-    return {"source": "a", "connectionId": keccak(connection_id)}
+    return {"source": "a" if is_mainnet else "b", "connectionId": keccak(connection_id)}
 
 
-def sign_l1_action(wallet, signature_types, signature_data, active_pool, nonce):
+def sign_l1_action(wallet, signature_types, signature_data, active_pool, nonce, is_mainnet):
     signature_types.append("address")
     signature_types.append("uint64")
     if active_pool is None:
@@ -117,7 +122,7 @@ def sign_l1_action(wallet, signature_types, signature_data, active_pool, nonce):
         signature_data.append(active_pool)
     signature_data.append(nonce)
 
-    phantom_agent = construct_phantom_agent(signature_types, signature_data)
+    phantom_agent = construct_phantom_agent(signature_types, signature_data, is_mainnet)
 
     data = {
         "domain": {
@@ -141,7 +146,63 @@ def sign_l1_action(wallet, signature_types, signature_data, active_pool, nonce):
         "primaryType": "Agent",
         "message": phantom_agent,
     }
+    return sign_inner(wallet, data)
 
+
+def sign_usd_transfer_action(wallet, message, is_mainnet):
+    data = {
+        "domain": {
+            "name": "Exchange",
+            "version": "1",
+            "chainId": 42161 if is_mainnet else 421613,
+            "verifyingContract": "0x0000000000000000000000000000000000000000",
+        },
+        "types": {
+            "UsdTransferSignPayload": [
+                {"name": "destination", "type": "string"},
+                {"name": "amount", "type": "string"},
+                {"name": "time", "type": "uint64"},
+            ],
+            "EIP712Domain": [
+                {"name": "name", "type": "string"},
+                {"name": "version", "type": "string"},
+                {"name": "chainId", "type": "uint256"},
+                {"name": "verifyingContract", "type": "address"},
+            ],
+        },
+        "primaryType": "UsdTransferSignPayload",
+        "message": message,
+    }
+    return sign_inner(wallet, data)
+
+
+def sign_agent(wallet, agent, is_mainnet):
+    data = {
+        "domain": {
+            "name": "Exchange",
+            "version": "1",
+            "chainId": 42161 if is_mainnet else 421613,
+            "verifyingContract": "0x0000000000000000000000000000000000000000",
+        },
+        "types": {
+            "Agent": [
+                {"name": "source", "type": "string"},
+                {"name": "connectionId", "type": "bytes32"},
+            ],
+            "EIP712Domain": [
+                {"name": "name", "type": "string"},
+                {"name": "version", "type": "string"},
+                {"name": "chainId", "type": "uint256"},
+                {"name": "verifyingContract", "type": "address"},
+            ],
+        },
+        "primaryType": "Agent",
+        "message": agent,
+    }
+    return sign_inner(wallet, data)
+
+
+def sign_inner(wallet, data):
     structured_data = encode_structured_data(data)
     signed = wallet.sign_message(structured_data)
     return {"r": to_hex(signed["r"]), "s": to_hex(signed["s"]), "v": signed["v"]}
